@@ -16,6 +16,7 @@ import socket
 import netaddr
 import unittest
 import ctypes
+import shutil
 
 try:
     from dbusmock import DBusTestCase
@@ -44,9 +45,111 @@ os.environ["GSETTINGS_BACKEND"] = "memory"
 os.dup2(sys.stdout.fileno(), sys.stderr.fileno())
 
 
+class WifiAuthentication():
+    def __init__(self):
+        self.wpa_settings = None
+        self.wpa_eap_settings = None
+        self.setting_name = None
+        self.needed_secrets = None
+
+        self.tmpdir = '/tmp/hostapd'
+        shutil.rmtree(self.tmpdir, ignore_errors=True)
+
+    @property
+    def auth_settings(self):
+        return self.wpa_settings
+
+    @property
+    def auth_eap_settings(self):
+        return self.wpa_eap_settings
+
+
+class WifiAuthenticationWPAPSK(WifiAuthentication):
+    def __init__(self, psk):
+        super().__init__()
+        self.wpa_settings = NM.SettingWirelessSecurity.new()
+        self.wpa_settings.set_property(NM.SETTING_WIRELESS_SECURITY_KEY_MGMT, "wpa-psk")
+        self.wpa_settings.set_property(NM.SETTING_WIRELESS_SECURITY_PSK, psk)
+
+        self.setting_name = NM.SETTING_WIRELESS_SECURITY_SETTING_NAME
+        self.needed_secrets = [NM.SETTING_WIRELESS_SECURITY_PSK]
+
+
+class WifiAuthenticationWPAEAP(WifiAuthentication):
+    def __init__(self, modes, phase2, identity, password, client_cert=False):
+        super().__init__()
+
+        self.make_certs()
+
+        self.wpa_settings = NM.SettingWirelessSecurity.new()
+        self.wpa_settings.set_property(NM.SETTING_WIRELESS_SECURITY_KEY_MGMT, "wpa-eap")
+
+        self.wpa_eap_settings = NM.Setting8021x.new()
+        self.wpa_eap_settings.set_property(NM.SETTING_802_1X_EAP, modes)
+
+        self.wpa_eap_settings.set_property(NM.SETTING_802_1X_IDENTITY, identity)
+
+        if password:
+            self.wpa_eap_settings.set_property(NM.SETTING_802_1X_PASSWORD, password)
+
+        if phase2 == 'tls':
+            self.wpa_eap_settings.set_property(NM.SETTING_802_1X_PHASE2_AUTHEAP, phase2)
+        else:
+            self.wpa_eap_settings.set_property(NM.SETTING_802_1X_PHASE2_AUTH, phase2)
+
+        if client_cert:
+            # Certificate paths must start with file:// and be null terminated
+            # See src/libnmc-setting/settings-docs.h
+            self.wpa_eap_settings.set_property(
+                    NM.SETTING_802_1X_CA_CERT,
+                    GLib.Bytes(f'file://{self.tmpdir}/pki/ca.crt\0'.encode()))
+            self.wpa_eap_settings.set_property(
+                    NM.SETTING_802_1X_CLIENT_CERT,
+                    GLib.Bytes(f'file://{self.tmpdir}/pki/issued/client.crt\0'.encode()))
+            self.wpa_eap_settings.set_property(
+                    NM.SETTING_802_1X_PRIVATE_KEY,
+                    GLib.Bytes(f'file://{self.tmpdir}/pki/private/client.key\0'.encode()))
+            self.wpa_eap_settings.set_property(
+                    NM.SETTING_802_1X_PRIVATE_KEY_PASSWORD, 'passw0rd')
+
+        self.setting_name = NM.SETTING_802_1X_SETTING_NAME
+        self.needed_secrets = [NM.SETTING_802_1X_PASSWORD, NM.SETTING_802_1X_PRIVATE_KEY_PASSWORD]
+
+    def make_certs(self):
+        os.mkdir(self.tmpdir)
+
+        with open(f'{self.tmpdir}/hostapd.eap_user', 'w') as f:
+            f.write('''* PEAP,TLS
+
+"account1" MSCHAPV2 "password1" [2]
+"account2" MSCHAPV2 "password2" [2]
+''')
+
+        # Create a CA, server and client certificates protected by passw0rd and the DH parameters
+        create_ca_script = """/usr/share/easy-rsa/easyrsa init-pki
+EASYRSA_BATCH=1 /usr/share/easy-rsa/easyrsa build-ca nopass
+EASYRSA_PASSOUT=pass:passw0rd EASYRSA_BATCH=1 /usr/share/easy-rsa/easyrsa build-server-full server
+EASYRSA_PASSOUT=pass:passw0rd EASYRSA_BATCH=1 /usr/share/easy-rsa/easyrsa build-client-full client
+/usr/share/easy-rsa/easyrsa gen-dh
+"""
+
+        with open(f'{self.tmpdir}/make_certs.sh', 'w') as f:
+            f.write(create_ca_script)
+
+        cmd = ['bash', 'make_certs.sh']
+        subprocess.call(cmd, stdout=subprocess.DEVNULL,
+                        stderr=subprocess.DEVNULL, cwd=self.tmpdir)
+
+
 class NetworkManagerTest(network_test_base.NetworkTestBase):
     """Provide common functionality for NM tests"""
 
+    def filtered_active_connections(self) -> list:
+        # Ignore the 'lo' connection, active since NM 1.42:
+        # https://networkmanager.dev/blog/networkmanager-1-42/#managing-the-loopback-interface
+        active_connections = [c for c in self.nmclient.get_active_connections() if c.get_id() != 'lo']
+        return active_connections
+
     def start_nm(self, wait_iface=None, auto_connect=True):
         """Start NetworkManager and initialize client object
 
@@ -64,6 +167,7 @@ class NetworkManagerTest(network_test_base.NetworkTestBase):
             "/etc/NetworkManager",
             "/var/lib/NetworkManager",
             "/run/NetworkManager",
+            "/etc/netplan",
         ]:
             subprocess.check_call(["mount", "-n", "-t", "tmpfs", "none", d])
             self.addCleanup(subprocess.call, ["umount", d])
@@ -106,6 +210,7 @@ class NetworkManagerTest(network_test_base.NetworkTestBase):
         argv += os.environ.get("NM_TEST_DAEMON_ARGS", "").strip().split()
 
         p = subprocess.Popen(argv, stdout=f_log, stderr=subprocess.STDOUT)
+        network_test_base.wait_nm_online()
         # automatically terminate process at end of test case
         self.addCleanup(p.wait)
         self.addCleanup(p.terminate)
@@ -253,11 +358,11 @@ class NetworkManagerTest(network_test_base.NetworkTestBase):
 
         return self.nmdev_w.get_access_points()[0]
 
-    def connect_to_ap(self, ap, secret, ipv6_mode, ip6_privacy):
+    def connect_to_ap(self, ap, auth_settings, ipv6_mode, ip6_privacy):
         """Connect to an NMAccessPoint.
 
-        secret should be None for open networks, and a string with the password
-        for WEP/WPA.
+        auth_settings should be None for open networks and an instance of
+        WifiAuthentication for WEP/WPA-PSK/WPA-EAP.
 
         ip6_privacy is a NM.SettingIP6ConfigPrivacy flag.
 
@@ -277,13 +382,10 @@ class NetworkManagerTest(network_test_base.NetworkTestBase):
         # map of a map of gpointers to gpointers which is too much for PyGI
         partial_conn = NM.SimpleConnection.new()
         partial_conn.add_setting(NM.SettingIP4Config(method=ip4_method))
-        if secret:
-            partial_conn.add_setting(NM.SettingWirelessSecurity.new())
-            # FIXME: needs update for other auth types
-            partial_conn.update_secrets(
-                NM.SETTING_WIRELESS_SECURITY_SETTING_NAME,
-                GLib.Variant("a{sv}", {"psk": GLib.Variant("s", secret)}),
-            )
+        if auth_settings:
+            partial_conn.add_setting(auth_settings.auth_settings)
+            if isinstance(auth_settings, WifiAuthenticationWPAEAP):
+                partial_conn.add_setting(auth_settings.auth_eap_settings)
         if ip6_privacy is not None:
             partial_conn.add_setting(
                 NM.SettingIP6Config(ip6_privacy=ip6_privacy, method=ip6_method)
@@ -337,16 +439,15 @@ class NetworkManagerTest(network_test_base.NetworkTestBase):
 
         # verify need_secrets()
         needed_secrets = conn.need_secrets()
-        if secret is None:
+        if auth_settings is None:
             self.assertEqual(needed_secrets, (None, []))
         else:
             self.assertEqual(
-                needed_secrets[0], NM.SETTING_WIRELESS_SECURITY_SETTING_NAME
+                needed_secrets[0], auth_settings.setting_name
             )
             self.assertEqual(type(needed_secrets[1]), list)
             self.assertGreaterEqual(len(needed_secrets[1]), 1)
-            # FIXME: needs update for other auth types
-            self.assertIn(needed_secrets[1][0], [NM.SETTING_WIRELESS_SECURITY_PSK])
+            self.assertIn(needed_secrets[1][0], auth_settings.needed_secrets)
 
         # we are usually ACTIVATING at this point; wait for completion
         # TODO: 5s is not enough, argh slow DHCP client
@@ -533,7 +634,7 @@ wpa_passphrase=12345678
             % SSID,
             None,
             54000,
-            "12345678",
+            WifiAuthenticationWPAPSK('12345678'),
         )
 
     def test_wpa2_ip4(self):
@@ -551,7 +652,7 @@ wpa_passphrase=12345678
             % SSID,
             None,
             54000,
-            "12345678",
+            WifiAuthenticationWPAPSK('12345678'),
         )
 
     def test_wpa2_ip6(self):
@@ -569,10 +670,70 @@ wpa_passphrase=12345678
             % SSID,
             "ra-only",
             54000,
-            "12345678",
+            WifiAuthenticationWPAPSK('12345678'),
             ip6_privacy=NM.SettingIP6ConfigPrivacy.PREFER_TEMP_ADDR,
         )
 
+    def test_wpa_enterprise_authentication_ip4(self):
+        """WPA2 + 802.1x, 802.11g, IPv4"""
+
+        self.do_test(
+            """hw_mode=g
+channel=1
+ssid=%s
+auth_algs=1
+eap_server=1
+ieee8021x=1
+eapol_version=2
+wpa=2
+wpa_key_mgmt=WPA-EAP
+wpa_pairwise=TKIP
+rsn_pairwise=CCMP
+eap_user_file=/tmp/hostapd/hostapd.eap_user
+ca_cert=/tmp/hostapd/pki/ca.crt
+server_cert=/tmp/hostapd/pki/issued/server.crt
+private_key=/tmp/hostapd/pki/private/server.key
+private_key_passwd=passw0rd
+dh_file=/tmp/hostapd/pki/dh.pem
+ctrl_interface=/var/run/hostapd
+ctrl_interface_group=0
+"""
+            % SSID,
+            None,
+            54000,
+            WifiAuthenticationWPAEAP(['peap'], 'mschapv2', 'account1', 'password1')
+        )
+
+    def test_wpa_enterprise_authentication_with_client_certificate_ip4(self):
+        """WPA2 + 802.1x with client certificate, 802.11g, IPv4"""
+
+        self.do_test(
+            """hw_mode=g
+channel=1
+ssid=%s
+auth_algs=1
+eap_server=1
+ieee8021x=1
+eapol_version=2
+wpa=2
+wpa_key_mgmt=WPA-EAP
+wpa_pairwise=TKIP CCMP
+rsn_pairwise=TKIP CCMP
+eap_user_file=/tmp/hostapd/hostapd.eap_user
+ca_cert=/tmp/hostapd/pki/ca.crt
+server_cert=/tmp/hostapd/pki/issued/server.crt
+private_key=/tmp/hostapd/pki/private/server.key
+private_key_passwd=passw0rd
+dh_file=/tmp/hostapd/pki/dh.pem
+ctrl_interface=/var/run/hostapd
+ctrl_interface_group=0
+"""
+            % SSID,
+            None,
+            54000,
+            WifiAuthenticationWPAEAP(['tls'], 'tls', 'client', None, client_cert=True)
+        )
+
     @network_test_base.run_in_subprocess
     def test_rfkill(self):
         """shut down connection on killswitch, restore it on unblock"""
@@ -619,7 +780,7 @@ wpa_passphrase=12345678
         hostapd_conf,
         ipv6_mode,
         expected_max_bitrate,
-        secret=None,
+        auth_settings=None,
         ip6_privacy=None,
     ):
         """Actual test code, parameterized for the particular test case"""
@@ -635,15 +796,15 @@ wpa_passphrase=12345678
         # self.assertEqual(ap.get_flags(), )
 
         # should not auto-connect
-        self.assertEqual(self.nmclient.get_active_connections(), [])
+        self.assertEqual(self.filtered_active_connections(), [])
 
         # connect to that AP
-        (conn, active_conn) = self.connect_to_ap(ap, secret, ipv6_mode, ip6_privacy)
+        (conn, active_conn) = self.connect_to_ap(ap, auth_settings, ipv6_mode, ip6_privacy)
 
         # check NMActiveConnection object
         self.assertIn(
             active_conn.get_uuid(),
-            [c.get_uuid() for c in self.nmclient.get_active_connections()],
+            [c.get_uuid() for c in self.filtered_active_connections()],
         )
         self.assertEqual(
             [d.get_udi() for d in active_conn.get_devices()], [self.nmdev_w.get_udi()]
@@ -653,7 +814,7 @@ wpa_passphrase=12345678
         wireless_setting = conn.get_setting_wireless()
         self.assertEqual(wireless_setting.get_ssid().get_data(), SSID.encode())
         self.assertEqual(wireless_setting.get_hidden(), False)
-        if secret:
+        if auth_settings:
             self.assertEqual(
                 conn.get_setting_wireless_security().get_name(),
                 NM.SETTING_WIRELESS_SECURITY_SETTING_NAME,
@@ -762,11 +923,11 @@ Logs are in '%s'. When done, exit the shell.
         if auto_connect:
             # ethernet should auto-connect quickly without an existing defined connection
             self.assertEventually(
-                lambda: len(self.nmclient.get_active_connections()) > 0,
+                lambda: len(self.filtered_active_connections()) > 0,
                 "timed out waiting for active connections",
                 timeout=100,
             )
-            active_conn = self.nmclient.get_active_connections()[0]
+            active_conn = self.filtered_active_connections()[0]
         else:
             # auto-connection was disabled, set up manual connection
             partial_conn = NM.SimpleConnection.new()
@@ -829,7 +990,7 @@ Logs are in '%s'. When done, exit the shell.
         # check NMActiveConnection object
         self.assertIn(
             active_conn.get_uuid(),
-            [c.get_uuid() for c in self.nmclient.get_active_connections()],
+            [c.get_uuid() for c in self.filtered_active_connections()],
         )
         self.assertEqual(
             [d.get_udi() for d in active_conn.get_devices()], [self.nmdev_e.get_udi()]
@@ -854,28 +1015,26 @@ class Hotplug(NetworkManagerTest):
     """In these tests APs are set up while NM is already running"""
 
     @network_test_base.run_in_subprocess
-    @unittest.expectedFailure
     def test_auto_detect_ap(self):
         """new AP is being detected automatically within 30s"""
 
-        self.setup_ap("hw_mode=b\nchannel=1\nssid=" + SSID, None)
         self.start_nm()
+        self.setup_ap("hw_mode=b\nchannel=1\nssid=" + SSID, None)
         ap = self.wait_ap(timeout=300)
         # get_ssid returns a byte array
-        self.assertEqual(ap.get_ssid(), SSID.encode())
+        self.assertEqual(ap.get_ssid().get_data(), SSID.encode())
         self.assertEqual(self.nmdev_w.get_active_access_point(), None)
 
     @network_test_base.run_in_subprocess
-    @unittest.expectedFailure
     def test_auto_detect_eth(self):
         """new eth router is being detected automatically within 30s"""
 
         self.start_nm()
         self.setup_eth(None)
         self.assertEventually(
-            lambda: len(self.nmclient.get_active_connections()) > 0, timeout=300
+            lambda: len(self.filtered_active_connections()) > 0, timeout=300
         )
-        active_conn = self.nmclient.get_active_connections()[0]
+        active_conn = self.filtered_active_connections()[0]
 
         self.assertEventually(
             lambda: active_conn.get_state() == NM.ActiveConnectionState.ACTIVATED,
